--***************************************************************************
-- Copyright (c) 1994 - Ohio Board of Regents and
--                    - University of Cincinnati, Cincinnati, OH.
--                    - Sidhartha Mohanty,
--                    - Department of Electrical & Computer Engineering,
--                    - Sept, 1994.
--***************************************************************************
--
-- Title:               Design library for Queueing Model. Version 0.0.
--
-- File:                [SUITE: vhdl/project/queue] queue.vhdl
-- Created:             Sun Sept 11 09:42:32 1994
-- Author:              Sidhartha Mohanty <smohanty@thor.ece.uc.edu>
--
-- Keywords:
-- LRM_version:         IEEE Std. P1076-1992/C
--
-- Objective:
--
-- Usage:
-- Analyze after packages.
--
-- Algorithm:
-- (None)
-- 
-- References:
-- 
-- Result:
-- 
-- History:
-- 
-- Comments:
-- 
--***************************************************************************

library queue;
use queue.queue_pkg.all;
use std.textio.all;

entity queue is

  generic (queue_length      : integer          := 1;
           no_initial_tokens : integer          := 0;
	   name 	     : string           := "Q1";
	   discipline        : queue_discipline := FIFO;
	   debug_control     : debug            := none;
           file_name         : string           := "";
	   report_freq 	     : time 	        := 10 fs);

  port (in_arc  : inout arc bus;  out_arc : inout arc bus);

end;

architecture behavior of queue is

  subtype queue_buff is token_array (1 to queue_length);
  signal number_tokens_held     : num_tokens_held := (no_initial_tokens, 0);
  signal number_changed		: tokens_change := 0;
  signal number_tokens_outgoing : integer := 0;  
  signal queue_arr 		: queue_buff;
  signal overflow		: flow_cntrl := '0';
  signal underflow		: flow_cntrl := '0';
  signal tail_change		: tail_ptr := 0;
  signal tail 			: integer := no_initial_tokens + 1;
  signal id 			: integer := 1;
  file outfile : text is out file_name;
begin

initialize : process

   variable integer_line : line;
   variable integer_id   : integer := 1;

  begin

    in_arc <= null;
    out_arc <= null;

    assert not std.textio.endfile(std.textio.input)
	report "(queue.vhdl) Number of entities in Queueing Model exceed number of integer ids. Increase the number."
	severity failure;

    readline(std.textio.input, integer_line);

    assert integer_line'length >= 0
	report "(queue.vhdl) Missing integer in identifiers file."
	severity failure;

    read(integer_line, integer_id);
    id <= integer_id;

    wait for 0 fs;

    wait;

end process;

enqueue : process 
    variable number_enqueued        : integer := 0;
    variable number_tokens_incoming : integer := 0;
    variable mesg : line;

  begin
  
    in_arc <= null;
    for i in 1 to no_initial_tokens loop
      queue_arr(i) <= (id, sending, TRUE, now);
    end loop;
    wait for 0 fs;

    loop 
    	wait on in_arc until (in_arc.handshake = sending);

  	if (debug_control = handshaking or 
  	    debug_control = summary) then 
  	  print_handshake_message ("Queue: ", name," received handshake signal: ", 
  		  		   in_arc.handshake, " from component: ", 
  				   in_arc.id, " at Time: ", now);
  	end if;
	number_tokens_incoming := number_tokens_incoming + 1;
        if ((number_tokens_held.held + 1) > queue_length) then
	   overflow <= '1';
           in_arc <= (id, not_ready, FALSE, now);
    	   assert ((number_tokens_held.held + 1) <= queue_length)
    	    report "Queue overflow."
    	    severity failure;
        else
	   assert FALSE
	     report "Enqueuing tokens"
	     severity note;
           if (debug_control = trace_tokens or 
 	       debug_control = summary) then 
           print_queue_message ("Queue: ", name, " enqueueing tokens from: ", 
  				in_arc.id, " at Time: ", now);
  	   end if;
           number_enqueued := number_enqueued + 1;
           write(mesg, string'("Total tokens enqueued at Queue: "));
           write(mesg, name);
           write(mesg, string'(" at Time: "));
           write(mesg, now);
           write(mesg, string'(" = "));
           write(mesg, number_enqueued);
           writeline(outfile, mesg);
           
    	   print_token_message ("Total tokens enqueued at Queue ",
  		    	        " at Time: ", now, " = ", number_enqueued);
	     if (tail = queue_length + 1) then 
		 tail_change <= - queue_length;
                 wait for 0 fs;
		 tail_change <= 0;
		 wait for 1 fs;
             end if; 
             queue_arr(tail) <= in_arc;
	     wait for 0 fs;
	     tail_change <= + 1;
	     wait for 0 fs;
	     tail_change <= 0;
	     wait for 0 fs;
	     underflow <= '0';
    	     number_changed <= + 1;
             wait for 0 fs;
	     number_changed <= 0;
             wait for 2 fs;
             if number_tokens_held.held < queue_length then
    	        in_arc <= (id, accepted, FALSE, now);
                wait for 0 fs;
	     else
		wait until number_tokens_held.held < queue_length;
		in_arc <= (id, accepted, FALSE, now);
                wait for 0 fs;
	     end if;
  	     if (debug_control = handshaking or 
  	         debug_control = summary) then 
  	       print_handshake_message ("Queue: ", name, " sent handshake signal: ", 
  				        in_arc.handshake, " to component: ", 
  				        in_arc.id, " at Time: ", now);
               end if;
	     in_arc <= null;
             wait for 0 fs;

        end if;
    end loop;
end process;

dequeue : process 

  variable head 	   : integer := 1;
  variable index 	   : integer := 1;
  variable number_dequeued : integer := 0;
  variable mesg : line;

  begin

    out_arc <= null;
    wait for 1 fs;  -- For the initializations to take place.

    loop
    
    	if (number_tokens_held.held >= 1) then
 	   assert FALSE
	     report "Dequeing tokens"
	     severity note;
  	   if (debug_control = trace_tokens or 
  	       debug_control = summary) then 
  	   print_queue_message ("Queue: ", name, " dequeueing token: ", 
  				queue_arr(head).id, " at Time: ", now);
  	   end if;
	   number_tokens_outgoing <= number_tokens_outgoing + 1;
           number_dequeued := number_dequeued + 1;
           write(mesg, string'("Total tokens dequeued at Queue: "));
           write(mesg, name);
           write(mesg, string'(" at Time: "));
           write(mesg, now);
           write(mesg, string'(" = "));
           write(mesg, number_dequeued);
           writeline(outfile, mesg);
           
  	   print_token_message ("Total tokens dequeued by Queue ",
  	                        " at Time: ", now, " = ", number_dequeued);
           
           if (discipline = FIFO) then
  	     print_formatted_message("Queue Discipline is FIFO");
             if (head = queue_length + 1) then head := 1; end if;
             out_arc <= queue_arr(head);
	     head := head + 1;
	     overflow <= '0'; 
    	     number_changed <= - 1;
             wait for 0 fs;
	     number_changed <= 0;
	     wait for 0 fs;
	     wait on out_arc until (out_arc.handshake = accepted);
  	     if (debug_control = handshaking or 
  	         debug_control = summary) then 
  	     print_handshake_message ("Queue: ", name, " received handshake signal: ", 
  				      out_arc.handshake, " from component: ", 
  				      out_arc.id, " at Time: ", now);
               end if;
	     out_arc <= null;
	     wait for 3 fs;
	   elsif (discipline = LIFO) then
  	     print_formatted_message("Queue Discipline is LIFO");
	     index := tail - 1;
	     out_arc <= queue_arr(index);
	     tail_change <= - 1;
	     wait for 0 fs;
	     tail_change <= 0;
	     wait for 0 fs;
	     overflow <= '0'; 
    	     number_changed <= - 1;
             wait for 0 fs;
	     number_changed <= 0;
	     wait for 0 fs;
	     wait on out_arc until (out_arc.handshake = accepted);

  	     if (debug_control = handshaking or 
  	         debug_control = summary) then 
  	       print_handshake_message ("Queue: ", name, " received handshake signal: ", 
  				        out_arc.handshake, " from component: ", 
  				        out_arc.id, " at Time: ", now);
               end if;
	     out_arc <= null;
	     wait for 3 fs;
           end if;
    	else
	    underflow <= '1';
	    assert false
	      report "Underflow: No tokens in Queue"
	      severity warning;
    	    wait until (number_tokens_held.held >= 1);
    	end if;
    end loop;
end process;    

count : process
 begin
   loop
     wait on number_changed;
     number_tokens_held <= (number_tokens_held.held + 
						 number_changed, 0);
   end loop;
end process;

count_tail : process
 begin
   loop
     wait on tail_change;
     tail <= tail + tail_change;  
   end loop;
end process;

stats : process
  
  variable avg_queue_len : integer;
  variable L : line;

 begin
  loop
    wait for 10 fs;
    avg_queue_len := number_tokens_held.held;
    if (debug_control = handshaking or 
  	         debug_control = summary) then
    write(L, string'("Average Queue Length (over a period of 10 fs) Qi(N): "));
    write(L, now - 10 fs);
    write(L, string'(" & "));
    write(L, now);
    write(L, string'(" = "));
    write(L, avg_queue_len);
    writeline(outfile, L);
    writeline(std.textio.output, L);
    end if;
  end loop;
end process;
  
end behavior;
